"""Useful general methods."""
import logging
import pandas as pd

logger = logging.getLogger(__name__)


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def export_table(results: pd.DataFrame, file_path: str):
    """Export results dataframe to path."""
    extension = file_path.split(".")[-1]
    acceptable_file_types = ("csv", "xlsx", "tsv")

    if extension not in acceptable_file_types:
        raise ValueError("Results output path must end with 'xlsx', 'csv', or 'tsv'")

    if extension == "xlsx":
        results.to_excel(file_path, index=False)

    elif extension in ("csv", "tsv"):
        separator = "," if extension == "csv" else "\t"
        results.to_csv(file_path, sep=separator, index=False)

    logger.info(f"Results written to {file_path}")



"""Default settings."""

import os
import logging
from pathlib import Path

from ebel_rest import connect
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy_utils import create_database, database_exists


logger = logging.getLogger(__name__)

# Graphstore credentials
db = 'pharmacome'
db_user = 'guest'
db_password = 'guest'
db_server = 'https://graphstore.scai.fraunhofer.de'

connect(db_user, db_password, db_server, db, print_url=False)

# Similar diseases to Alzheimer's
SIMILAR_DISEASES = ["Parkinson Disease", "Amyotrophic Lateral Sclerosis", "Neurodegenerative Diseases",
                    "Huntington Disease"]

# Default Directory Paths
HOME = str(Path.home())
PROJECT_DIR = os.path.join(HOME, ".drugintfinder")
LOG_DIR = os.path.join(PROJECT_DIR, "logs")
CACHE_DIR = os.path.join(PROJECT_DIR, "cache")

os.makedirs(LOG_DIR, exist_ok=True)
os.makedirs(CACHE_DIR, exist_ok=True)

# Logging Configuration
LOG_FILE_PATH = os.path.join(LOG_DIR, "drugintfinder.log")
logging.basicConfig(filename=LOG_FILE_PATH,
                    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

# SQLite
DEFAULT_DB_PATH = os.path.join(PROJECT_DIR, "drugintfinder.db")
CONN_ENV = os.getenv('DIF_DB_PATH')
CONN_PATH = CONN_ENV if CONN_ENV else DEFAULT_DB_PATH
CONN = f"sqlite:///{CONN_PATH}"
if not database_exists(CONN):
    create_database(CONN)

engine = create_engine(CONN, convert_unicode=True)
session = sessionmaker(bind=engine)



"""A pipeline for identifying the most notable druggable interactors of a target within a e(BE:L) Knowledge Graph.

This package is designed for identifying the direct interactors of a specified node within an e(BE:L) generated
Knowledge Graph. The starting node can be filtered by given protein modifications and the type of interactors
(e.g. gene, RNA, protein) can also be controlled. Additioanlly, this package also can identify those interactors
that are druggable and provide a table of useful statistics about each drug/target combination that can be used for
ranking these pairs.

To get started with the druggable interactor finder, you can install it using:
``pip install drugintfinder``
"""

from drugintfinder import defaults

__author__ = """Bruce Schultz"""
__email__ = 'bruce.schultz@scai.fraunhofer.de'
__version__ = '0.2.3'



"""Calculate the ranking of the hits."""

import json
import logging
import requests
import pandas as pd

from tqdm import tqdm
from typing import Optional
from ebel_rest import query as rest_query
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.exc import MultipleResultsFound

from drugintfinder.constants import PATENTS, INTERACTORS, PRODUCTS, IDENTIFIERS, NEGREG, POSREG, SYNERGY, POINTS, \
    DAC, ACTION_MAPPER, CT_MAPPER, CLINICAL_TRIALS, TRIALS, IN_COUNT, OUT_COUNT, PUBCHEM_BIOASSAY_API, UNIPROT_ID, \
    TIC, ASSOCIATED_PATHWAYS, TARGET_COUNT
from drugintfinder.finder import InteractorFinder
from drugintfinder.defaults import session, SIMILAR_DISEASES
from drugintfinder.models import Trials, TargetStats, Patents, Products, Drugs, BioAssays

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class Ranker:
    """Rank and annotate druggable interactors identified using the InteractorFinder class."""

    def __init__(
            self,
            name: str,
            node_type: str = "protein",
            pmods: list = None,
            penalty: int = -1,
            reward: int = 1,
            disease_keyword: str = "Alzheimer Disease",
            similar_diseases: list = SIMILAR_DISEASES,
            print_sql: bool = False
    ):
        """Init for Ranker.

        Should be initialized with an InteractorFinder object that has results saved.
        """
        self.__session = session()
        self.node_name = name
        self.pmods = pmods
        self.disease = disease_keyword
        self.similar_diseases = similar_diseases

        self.__finder = InteractorFinder(
            node_name=name, node_type=node_type, pmods=pmods, neighbor_edge_type='causal', print_sql=print_sql
        )
        self.__finder.druggable_interactors()
        self.table = self.__finder.results
        if self.table is None:
            raise ValueError(f"No druggable interactors were found for {self.node_name} + {self.pmods}")

        self.__penalty = penalty
        self.__reward = reward

        self.interactor_metadata = {int_name: dict() for int_name in self.interactors}

        self.__cached_drugs = self.__session.query(Drugs.id,
                                                   Drugs.drug_name,
                                                   Drugs.num_targets,
                                                   Drugs.clinical_trials)\
            .filter(Drugs.drug_name.in_(self.interactor_drugs))\
            .all()

        target_mapper, drug_ids, cts = self.__parse_cached_data()
        self.target_count_mapper = target_mapper
        self.relevant_drug_row_ids = drug_ids
        self.relevant_cts = cts

        self.drug_scores = self.__generate_ranking_dict()
        self.drug_metadata = self.__compile_drug_metadata()

        self.ct_summary = pd.DataFrame()

    def __parse_cached_data(self) -> tuple:
        target_mapper = dict()
        drug_ids = []
        cts = dict()
        for drug_entry in self.__cached_drugs:
            drug_id, drug_name, targets, cts_str = drug_entry
            target_mapper[drug_name] = targets
            drug_ids.append(drug_id)
            cts[drug_name] = [trial_id for trial_id in cts_str.split("|") if trial_id]

        return target_mapper, drug_ids, cts

    @property
    def interactors(self) -> list:
        """Compile a list of all identified interactors."""
        return list(self.__finder.unique_interactors())

    @property
    def interactor_drugs(self) -> list:
        """Compile a list of all identified drugs."""
        return list(self.__finder.unique_drugs())

    @property
    def unique_drug_target_combos(self) -> set:
        """Create a list of tuples uniue combinations of drugs and their targets from the ranking results."""
        unique_pairs = set()
        pairs = self.__finder.drug_and_interactors().to_dict('records')

        for combo in pairs:
            unique_pairs.add((combo['drug'], combo['interactor_name']))

        return unique_pairs

    @property
    def dbid_drugname_mapper(self):
        """Map DrugBank IDs to their drug names."""
        return {data[IDENTIFIERS]['drugbank_id']: drug_name for drug_name, data in self.drug_metadata.items()}

    def rank(self):
        """Score drugs and their targets."""
        self.score_drugs()
        self.score_interactors()

    def score_drugs(self):
        """Parse raw drug metadata and calculate points for each ranking criteria."""
        self.score_drug_relationships()
        self.score_patents_and_products()
        self.score_cts()
        # self.score_homologs()

    def score_interactors(self):
        """Parse raw interactor metadata and calculate points for each ranking criteria."""
        self.count_bioassays()
        self.count_edges()

    def __compile_drug_metadata(self) -> dict:
        metadata = self.__generate_ranking_dict()
        logger.info("Compiling drug metadata")

        for r in self.table.to_dict('records'):
            drug_name = r['drug']
            db_id = r['drugbank_id']
            interactor_name = r['interactor_name']
            metadata[drug_name][IDENTIFIERS] = {'drugbank_id': db_id,
                                                'chembl_id': r['chembl_id'],
                                                'pubchem_id': r['pubchem_id']}

            # Parse relations
            if r['relation_type'] == 'regulates':
                continue

            elif interactor_name in metadata[drug_name][INTERACTORS]:
                metadata[drug_name][INTERACTORS][interactor_name]['relation_type'].add(r['relation_type'])

            else:
                rel_data = {'relation_type': {r['relation_type']},  # Collect rel types in set for later comparison
                            'actions': r['drug_rel_actions'].split("|") if r['drug_rel_actions'] else None}
                metadata[drug_name][INTERACTORS][interactor_name] = rel_data

        return metadata

    def __generate_ranking_dict(self) -> dict:
        """Compile a generic empty dict to fill with raw and scored data."""
        empty_dict = {drug_name: {PATENTS: dict(),
                                  PRODUCTS: dict(),
                                  IDENTIFIERS: dict(),
                                  INTERACTORS: dict(),
                                  CLINICAL_TRIALS: dict(),
                                  TARGET_COUNT: None,
                                  }
                      for drug_name in self.interactor_drugs}
        return empty_dict

    def score_patents_and_products(self):
        """Parse patent and product information and generates a score."""
        logger.info("Scoring patent and product information")

        relevant_patents = self.__session.query(Drugs.drug_name, Patents.expired) \
            .join(Drugs).filter(Patents.drug_id.in_(self.relevant_drug_row_ids)).all()

        relevant_products = self.__session.query(Drugs.drug_name, Products.has_approved_generic) \
            .join(Drugs).filter(Products.drug_id.in_(self.relevant_drug_row_ids)).all()

        patent_mapper = dict()
        for drug_name, expired in relevant_patents:
            if drug_name in patent_mapper:
                patent_mapper[drug_name].append(expired)
            else:
                patent_mapper[drug_name] = [expired]

        product_mapper = dict()
        for drug_name, approved_generic in relevant_products:
            if drug_name in product_mapper:
                product_mapper[drug_name].append(approved_generic)
            else:
                product_mapper[drug_name] = [approved_generic]

        for drug_name in self.drug_scores.keys():
            all_patents_expired = all(patent_mapper[drug_name]) if drug_name in patent_mapper else False
            approved_generic_available = any(product_mapper[drug_name]) if drug_name in product_mapper else False
            self.drug_scores[drug_name][PATENTS] = {'expired': all_patents_expired,
                                                    POINTS: self.__reward if all_patents_expired else self.__penalty}
            self.drug_scores[drug_name][PRODUCTS] = {'has_approved_generic': approved_generic_available,
                                                     POINTS: self.__reward if approved_generic_available
                                                     else self.__penalty}

            try:
                self.drug_scores[drug_name][TARGET_COUNT] = self.target_count_mapper[drug_name]

            except KeyError:  # If drug not in cache then probably forgot to populate
                logger.error(f"{drug_name} not found in cache. Try populate.populate() again.")

    def score_homologs(self, tc_json: str, db_id_mapper: dict, tc_threshold: int = 0.95):
        """Produce a dictionary detailing structural homologs of every drugbank drug."""
        with open(tc_json, 'r') as mcsf:
            content = json.load(mcsf)
        homolog_scores = dict()
        for db_id, props in content.items():
            if 'similarities' in props and db_id in db_id_mapper:  # Check if in mapper else not in KG list
                drug_name = db_id_mapper[db_id]
                homolog_scores[drug_name] = {'homologs': {}, POINTS: self.__penalty}
                for db_id2, tc in props['similarities'].items():
                    if tc >= tc_threshold:
                        homolog_scores[drug_name]['homologs'][db_id2] = tc
                        homolog_scores[drug_name][POINTS] = self.__reward
        return homolog_scores

    @staticmethod
    def __check_target_interactor_contradictions(interactor_metadata: dict) -> bool:
        """Check whether there are contradicting relation types between the interactor and target."""
        tests = (('increases', 'decreases'),
                 ('increases', 'directly_decreases'),
                 ('decreases', 'directly_increases'),
                 ('directly_increases', 'directly_decreases'))

        relations = interactor_metadata['relation_type']
        contradictions = []
        for test in tests:
            contradiction_check = all(rel in relations for rel in test)
            contradictions.append(contradiction_check)

        return any(contradictions)

    @staticmethod
    def __check_drug_action_contradictions(drug_actions: set) -> bool:
        """Check if both 'positive_regulator' and 'negative_regulator' in the set of mapped drug/target relations."""
        contradiction = False
        if POSREG and NEGREG in drug_actions:
            contradiction = True
        return contradiction

    @staticmethod
    def __check_rel_synergy(drug_actions: set, int_rel: set) -> Optional[bool]:
        """Return True is drug/target relation and interactor/pTAU relation results in decrease of pTAU else False."""
        pos_int_rel = ['increases', 'directly_increases']
        neg_int_rel = ['decreases', 'directly_decreases']

        # TODO make generic so it works for any target
        # drug -inhibits-> interactor -decreases-> pTAU: BAD
        if NEGREG in drug_actions and any(rel in int_rel for rel in neg_int_rel):
            return False

        # drug -inhibits-> interactor -increases-> pTAU: GOOD
        elif NEGREG in drug_actions and any(rel in int_rel for rel in pos_int_rel):
            return True

        # drug -activates-> interactor -increases-> pTAU: BAD
        elif POSREG in drug_actions and any(rel in int_rel for rel in pos_int_rel):
            return False

        # drug -activates-> interactor -decreases-> pTAU: GOOD
        elif POSREG in drug_actions and any(rel in int_rel for rel in neg_int_rel):
            return True

        else:  # drug_actions = 'neutral'
            return None

    def score_drug_relationships(self):
        """Add the score to the drug/target pair.

        Score is based on whether it has information on the action (inhibitor/activator) and whether it makes
        sense based on the relationship of the interactor with target.

        Checks:
        * Whether there are contradicting edges between target/interactor (TIC) --> penalty
        * Whether there are contradicting drug actions between drug/interactor (DAC) --> penalty
        * If the drug action and target/interactor relationship synergize to produce desired effect (synergy)

        Returns
        -------
        checked_drug_rels: dict
            Keys are drug names, values are point values.
        """
        logger.info("Scoring drug relationships")
        for drug_name, metadata in self.drug_metadata.items():
            for target_name, ti_metadata in metadata[INTERACTORS].items():
                self.drug_scores[drug_name][INTERACTORS][target_name] = {
                    TIC: False,
                    DAC: False,
                    SYNERGY: None,
                    POINTS: 0
                }
                tic = self.__check_target_interactor_contradictions(ti_metadata)  # Target/interactor

                pts = 0
                if tic is True:  # There are contradicting edges between target and interactor
                    pts += self.__penalty
                    self.drug_scores[drug_name][INTERACTORS][target_name][TIC] = tic
                    self.drug_scores[drug_name][INTERACTORS][target_name][POINTS] = pts
                    continue  # No point in continuing since drug/target action and int/target rel can't be compared

                rels = ti_metadata['relation_type']  # imported as a set

                if ti_metadata['actions'] is None:
                    pts += self.__penalty

                else:
                    mapped_actions = {ACTION_MAPPER[rel] for rel in ti_metadata['actions']}
                    # First check if drug actions have contradictions
                    dac = self.__check_drug_action_contradictions(mapped_actions)
                    if dac is True:
                        self.drug_scores[drug_name][INTERACTORS][target_name][DAC] = dac
                        pts += self.__penalty

                    else:  # Compare synergy of drug actions with interactor/pTAU relations
                        synergy = self.__check_rel_synergy(mapped_actions, rels)
                        self.drug_scores[drug_name][INTERACTORS][target_name][SYNERGY] = synergy

                        if synergy is True:  # Good comparison
                            pts += self.__reward

                        else:
                            pts += self.__penalty

                self.drug_scores[drug_name][INTERACTORS][target_name][POINTS] = pts

    @staticmethod
    def __chunk_cts(trial_id_list: list, chunk_size: int = 100) -> list:
        """Create a generator of trial ID lists."""
        total_num_chunks = (len(trial_id_list) // chunk_size) + 1
        chunk_index = 0
        while chunk_index < total_num_chunks:
            start = chunk_index * chunk_size
            stop = (chunk_index + 1) * chunk_size
            yield trial_id_list[start:stop]
            chunk_index += 1

    def __build_ct_cache(self) -> dict:
        """Build a dictionary of metadata for each clinical trial relevant to the query."""
        cached_ct_data = dict()
        relevant_trial_ids = list({trial_id for id_list in self.relevant_cts.values() for trial_id in id_list})

        trial_id_chunks = self.__chunk_cts(relevant_trial_ids, chunk_size=400)
        for id_batch in trial_id_chunks:
            ct_results = self.__session.query(Trials.trial_id, Trials.conditions, Trials.trial_status)\
                .filter(Trials.trial_id.in_(id_batch)).all()

            for ct_entry in ct_results:
                trial_id, conditions, trial_status = ct_entry

                metadata = {'conditions': conditions.split(";"), 'trial_status': trial_status}
                cached_ct_data[trial_id] = metadata

        return cached_ct_data

    def __compile_ct_metadata(self) -> dict:
        """Compile the Clinical Trial data from the DB."""
        ct_mapper = dict()
        cached_ct_data = self.__build_ct_cache()

        for drug_name, trial_ids in self.relevant_cts.items():
            for trial_id in trial_ids:
                metadata = cached_ct_data[trial_id]

                if drug_name in ct_mapper:
                    ct_mapper[drug_name][trial_id] = metadata

                else:
                    ct_mapper[drug_name] = {trial_id: metadata}

        return ct_mapper

    def score_cts(self):
        """Score drugs based on involvement in a clinical trial.

        Returned dictionary contains points and relevant AD-associated CT information.
        """
        logger.info("Scoring Clinical Trial data")
        ct_mapper = self.__compile_ct_metadata()
        ct_summary_rows = []

        for drug_name, ct_metadata in ct_mapper.items():
            pts = self.__reward  # Default to reward unless changed

            self.drug_scores[drug_name][CLINICAL_TRIALS][TRIALS] = dict()
            for trial_id, trial_data in ct_metadata.items():
                ct_score = {'keyword_disease_investigated': False,
                            'trial_ongoing': False,
                            'similar_disease_investigated': False,
                            'conditions_investigated': trial_data['conditions']}

                if self.disease in trial_data['conditions']:  # Disease-associated CT
                    ct_score['keyword_disease_investigated'] = True
                    if trial_data['trial_status'] in CT_MAPPER['ongoing']:
                        ct_score['trial_ongoing'] = True
                        pts += self.__penalty

                else:  # Not associated with primary disease
                    if set(trial_data['conditions']) & set(self.similar_diseases):  # Trial for similar disease
                        ct_score['similar_disease_investigated'] = True
                        pts += self.__reward * 2  # Double points
                    else:
                        if trial_data['trial_status'] in CT_MAPPER['ongoing']:
                            ct_score['trial_ongoing'] = True
                            pts += self.__penalty

                self.drug_scores[drug_name][CLINICAL_TRIALS][TRIALS][trial_id] = ct_score
                ct_row = {'drug': drug_name, 'trial_id': trial_id, **ct_score}
                ct_summary_rows.append(ct_row)

            self.drug_scores[drug_name][CLINICAL_TRIALS][POINTS] = pts

        # Compile summary dataframe of CT information
        column_names = ["Drug", "Trial ID", "Keyword Disease Investigated", "Trial Ongoing",
                        "Similar Disease Investigated", "Conditions Investigated"]

        if ct_summary_rows:
            self.ct_summary = pd.DataFrame(ct_summary_rows)
            self.ct_summary.columns = column_names
        else:
            self.ct_summary = pd.DataFrame(columns=column_names)

    def count_associated_pathways(self) -> dict:
        """Go through each interactor and checks if it is associated with a KEGG or Pathway Commons pathway."""
        pathways = dict()
        for protein in tqdm(self.interactors, desc="Checking pathways"):
            r = rest_query.sql(ASSOCIATED_PATHWAYS.format(protein, protein)).data
            pathways[protein] = r[0]['count']

        return pathways

    @staticmethod
    def __query_db_edge_counts(symbol: str) -> Optional[dict]:
        """Obtain counts from SQLite DB for given gene symbol."""
        logger.info("Querying SQLite DB for edge counts")
        try:
            results = session().query(TargetStats).filter_by(symbol=symbol).one()

        except MultipleResultsFound as mrfe:
            print(mrfe)
            logger.warning(f"Multiple results found for {symbol}")
            results = None

        except NoResultFound:
            logger.info(f"No edge counts found for {symbol}")
            results = None

        if results:
            data = {'symbol': results.symbol,
                    'out_count': results.out_count,
                    'in_count': results.in_count,
                    'both_count': results.both_count}
            return data

    @staticmethod
    def __query_graphstore_edge_counts(symbol: str) -> dict:
        """Query the graphstore for the edge counts for given gene symbol."""
        in_count = rest_query.sql(IN_COUNT.format(symbol)).data[0]['number']
        out_count = rest_query.sql(OUT_COUNT.format(symbol)).data[0]['number']
        both_count = out_count + in_count

        counts = {'in_count': in_count, 'out_count': out_count, 'both_count': both_count, 'symbol': symbol}

        logger.info(f"Importing edge counts into DB for {symbol}")
        sess = session()
        edge_entry = TargetStats(**counts)
        sess.add(edge_entry)
        sess.commit()

        return counts

    def count_edges(self) -> dict:
        """Count the number of incoming, outgoing, and total edges for each interactor."""
        edge_counts = dict()
        logger.info("Gathering edge counts")
        for symbol in tqdm(self.interactor_metadata.keys(), total=len(self.interactor_metadata),
                           desc="Counting edges"):
            counts = self.__query_db_edge_counts(symbol)
            if not counts:
                counts = self.__query_graphstore_edge_counts(symbol)
            self.interactor_metadata[symbol]['edges'] = counts
            edge_counts[symbol] = counts

        return edge_counts

    def __get_bioassay_cache(self) -> dict:
        """Get BioAssay counts in DB and returns a dict - key: protein target symbol, val: bioassay count."""
        bioassay_counts = self.__session.query(BioAssays.target, BioAssays.num_assays).all()
        return {r[0]: r[1] for r in bioassay_counts}

    def count_bioassays(self) -> dict:
        """Query the PubChem API to obtain the number of available BioAssays for each gene symbol in the list.

        Returns
        -------
        dict
            Key is the gene symbol, value is the number of BioAssays available.
        """
        counts = self.__get_bioassay_cache()
        symbols_missing_data = set(self.interactors) - set(counts.keys())  # Create set of symbols not in cache

        if symbols_missing_data:
            for symbol in tqdm(symbols_missing_data, desc="Counting BioAssays for targets"):

                up_acc_results = rest_query.sql(UNIPROT_ID.format(symbol)).data

                if up_acc_results:
                    up_acc = up_acc_results[0]['uniprot.id']

                    filled = PUBCHEM_BIOASSAY_API.format(up_acc)
                    resp = requests.get(filled)
                    number_assays = len(resp.text.split("\n")) - 1  # Remove header
                    counts[symbol] = number_assays

                    row = BioAssays(target=symbol, num_assays=number_assays)
                    self.__session.add(row)

            self.__session.commit()

        for symbol in self.interactor_metadata.keys():
            self.interactor_metadata[symbol]['bioassays'] = counts[symbol]  # Number of assays

        return counts

    def __create_summary_table(self) -> pd.DataFrame:
        rows = []
        for drug_name, symbol in self.unique_drug_target_combos:
            drug_entry = self.drug_scores[drug_name]

            synergy_result = drug_entry[INTERACTORS][symbol][SYNERGY]
            if synergy_result is None:
                synergizes = "N/A"

            elif synergy_result is False:
                synergizes = "No"

            else:  # is True
                synergizes = "Yes"

            # Interactor metadata
            num_bioassays = self.interactor_metadata[symbol]['bioassays']
            num_total_edges = self.interactor_metadata[symbol]['edges']['both_count']

            # Drug metadata
            ongoing_patent = "Yes" if drug_entry[PATENTS]['expired'] is True else "No"
            has_generic = "Yes" if drug_entry[PRODUCTS]['has_approved_generic'] is True else "No"
            target_count = drug_entry[TARGET_COUNT]
            target_count_entry = "N/A" if target_count == -1 or not target_count else target_count

            # Compile
            row = {'Drug': drug_name,
                   'Target': symbol,
                   'Synergizes': synergizes,
                   'Number of BioAssays for Target': num_bioassays,
                   'Number of Causal Edges for Target': num_total_edges,
                   'Drug Patent Ongoing': ongoing_patent,
                   'Generic Version of Drug Available': has_generic,
                   'Number of Drug Targets': target_count_entry}
            rows.append(row)

        return pd.DataFrame(rows)

    def summarize(self, pivot: bool = False) -> pd.DataFrame:
        """Summarize the ranking results.

        Be default, the summarized table is compiled into rows of unique drug/targets. If `pivot` is enabled, the
        table instead focuses on the targets and the columns adjust accordingly.

        :param pivot: If True, drugs are compressed into a new column and table now focuses on targets.
        :return: pd.DataFrame os summary results.
        """
        summary_df = self.__create_summary_table()

        if pivot:
            num_drugs_mapper = summary_df.Target.value_counts()
            summary_df = summary_df.drop([
                'Drug',
                'Synergizes',
                'Drug Patent Ongoing',
                'Generic Version of Drug Available',
                'Number of Drug Targets'
            ], axis=1).drop_duplicates().reset_index(drop=True)
            summary_df["Number of Known Drugs"] = [num_drugs_mapper[target] for target in summary_df.Target]
            summary_df["BioAssay to Known Drug Ratio"] = \
                summary_df["Number of BioAssays for Target"] / summary_df["Number of Known Drugs"]
            summary_df = summary_df.sort_values(by="Number of BioAssays for Target", ascending=False)

        return summary_df



"""Methods for executing pipeline."""
import logging

import pandas as pd

from typing import Optional
from ebel_rest import query as rest_query

from drugintfinder.populate import populate
from drugintfinder.defaults import session, engine
from drugintfinder.models import General, Druggable
from drugintfinder.constants import INTERACTOR_QUERY, PURE_DRUGGABLE_QUERY, CAPSULE_DRUGGABLE_COMPLEX, \
    CAPSULE_DRUGGABLE_MODIFIED, EDGE_MAPPER

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class InteractorFinder:
    """Find the different types of interactors for a given node symbol and protein specification."""

    def __init__(
            self,
            node_name: str,
            node_type: str = 'protein',
            pmods: list = None,
            neighbor_edge_type: str = 'E',
            print_sql: bool = False
    ):
        """Init method for InteractorFinder class.

        Parameters
        ----------
        node_name : str
            Value of target node for which you want to find interactors for. This could be a gene symbol, disease name,
            or any other accepted BEL value.
        node_type : str
            Specify the type of node the target is. Defaults to "protein", but can be any BEL node type.
        pmods : list
            If the target node is a protein, protein modifications can be added to restrict searching to protein nodes
            with both the given name AND the given pmods. Can use BEL shorthand pmods (e.g. "pho", "ace", "me0") or
            pmod values (e.g. "protein phosphorylation").
        neighbor_edge_type : str
            The class of edges to consider when looking for interactors. Defaults to "E" which encompasses all edges
            types, but more specific classes can be given such as only BEL edge types ("bel"), causal edges "causal",
            or "increases"/"directly_increases".
        print_sql : bool
            If True, will print each query as it is executed in OrientDB.
        """
        populate()  # Populate tables if empty

        self.node_name = node_name.lower()
        self.node_type = node_type
        self.pmods = pmods
        self.edge = neighbor_edge_type
        self.print_sql = print_sql

        self.results = None

    def __len__(self):
        return len(self.results)

    def __query_graphstore(self, sql: str) -> Optional[pd.DataFrame]:
        if self.print_sql:
            print(sql)

        logger.warning("Querying graphstore for interactors - this may take awhile")
        results = rest_query.sql(sql).table

        if isinstance(results, str):
            logger.warning("No results!")
            return None

        else:
            return results

    def __query_db(self, druggable: bool = False):
        """Check if query results are stored in cache."""
        target = self.node_name.upper()  # Target symbol upper case for humans
        if druggable:
            rels = EDGE_MAPPER['causal']

        elif self.edge in EDGE_MAPPER:
            rels = EDGE_MAPPER[self.edge]

        else:
            rels = [self.edge]

        table = Druggable if druggable else General

        query = session().query(table).filter_by(target_symbol=target, target_type=self.node_type)
        results = pd.read_sql(query.statement, query.session.bind)

        # Filter by edges and pmods
        filtered_df = results[results['relation_type'].isin(rels)]
        if self.pmods:
            filtered_df = filtered_df[filtered_df['pmod_type'].isin(self.pmods)]

        return filtered_df if not filtered_df.empty else None

    def find_interactors(self) -> pd.DataFrame:
        """Return interactors of the target."""
        table = General.__tablename__
        cached_results = self.__query_db(druggable=False)
        if cached_results is not None and not cached_results.empty:
            self.results = cached_results

        else:
            sql = INTERACTOR_QUERY

            cols = ["target_species", "pmid", "pmc", "interactor_type", "interactor_name", "interactor_bel",
                    "relation_type", "target_bel", "target_type", "target_symbol", "pmod_type"]

            if self.node_type != 'protein' or not self.pmods:
                sql = sql.replace('MATCH {{class:pmod, as:pmod{}}}<-has__pmod-', 'MATCH')
                formatted_sql = sql.format(self.node_type, self.node_name, self.edge)

            else:
                if 'all' in self.pmods:
                    pmod_condition = "type != '' or name != ''"
                else:
                    pmod_condition = f"type in {self.pmods}"

                pmod_string = f", WHERE:({pmod_condition})"

                if 'pho' in self.pmods or 'all' in self.pmods:
                    pmod_string = pmod_string.replace(")", " OR name like '%phosphorylat%')")
                formatted_sql = sql.format(pmod_string, self.node_type, self.node_name, self.edge)

            df_results = self.__query_graphstore(formatted_sql)

            if df_results is not None:
                self.results = df_results[cols]
                self.results['target_species'] = self.results['target_species'].fillna(0).astype(int)

                logger.info(f"Importing {table} results for {self.node_name} into SQLite DB")
                self.results.to_sql(table, if_exists="append", con=engine, index=False)

        return self.results

    def druggable_interactors(self) -> pd.DataFrame:
        """Return all druggable interactors of the target.

        Requires specialized queries and therefore is separate from `find_interactors`.
        """
        table = Druggable.__tablename__

        cached_results = self.__query_db(druggable=True)
        if cached_results is not None and not cached_results.empty:
            self.results = cached_results

        else:
            cols = ['drug', 'capsule_interactor_type', 'capsule_interactor_bel', 'interactor_bel', 'interactor_type',
                    'interactor_name', 'relation_type', 'target_bel', 'target_symbol', 'target_type',
                    'pmid', 'pmc', 'rel_pub_year', 'rel_rid', 'drug_rel_rid', 'drug_rel_actions',
                    'drugbank_id', 'chembl_id', 'pubchem_id', 'pmod_type']

            if self.node_type != 'protein' or not self.pmods:
                pure_query = PURE_DRUGGABLE_QUERY.replace('MATCH {{class:pmod, as:pmod{}}}<-has__pmod-', 'MATCH')
                capsule_query_1 = CAPSULE_DRUGGABLE_MODIFIED.replace(
                    'MATCH {{class:pmod, as:pmod{}}}<-has__pmod-', 'MATCH'
                )
                capsule_query_2 = CAPSULE_DRUGGABLE_COMPLEX.replace(
                    'MATCH {{class:pmod, as:pmod{}}}<-has__pmod-', 'MATCH'
                )
                formatted_pure_sql = pure_query.format(self.node_type, self.node_name)
                formatted_capsule_sql_1 = capsule_query_1.format(self.node_type, self.node_name)
                formatted_capsule_sql_2 = capsule_query_2.format(self.node_type, self.node_name)

            else:
                if 'all' in self.pmods:
                    pmod_condition = "type != '' or name != ''"
                else:
                    pmod_condition = f"type in {self.pmods}"

                pmod_string = f", WHERE:({pmod_condition})"

                if 'pho' in self.pmods or 'all' in self.pmods:
                    pmod_string = pmod_string.replace(")", " OR name like '%phosphorylat%')")

                # Drugs only for humans so only check one
                formatted_pure_sql = PURE_DRUGGABLE_QUERY.format(self.node_type, self.node_name)
                formatted_capsule_sql_1 = CAPSULE_DRUGGABLE_MODIFIED.format(self.node_type, self.node_name)
                formatted_capsule_sql_2 = CAPSULE_DRUGGABLE_COMPLEX.format(self.node_type, self.node_name)

            logger.info("Querying database...")

            pure_results = self.__query_graphstore(sql=formatted_pure_sql)
            capsule_results_1 = self.__query_graphstore(sql=formatted_capsule_sql_1)
            capsule_results_2 = self.__query_graphstore(sql=formatted_capsule_sql_2)

            if any([pure_results, capsule_results_1, capsule_results_2]):  # Only need one to have results
                df_concat = pd.concat(
                    [pure_results, capsule_results_1, capsule_results_2], axis=0
                ).reindex(columns=cols)
                self.results = df_concat[cols].drop_duplicates()
                self.results["drug_rel_actions"] = self.results["drug_rel_actions"].str.join("|")

                logger.info(f"Importing {table} results for {self.node_name} into SQLite DB")
                self.results.to_sql(table, if_exists="append", con=engine, index=False)

        return self.results

    def drug_and_interactors(self):
        """Return a list of interactors and the drugs that affect them."""
        if self.results is not None and 'drug' in self.results.columns:
            return self.results[['drug', 'interactor_name']]

    def unique_interactors(self):
        """Return a list of unique interactors found in the results dataframe."""
        if self.results is not None:
            return tuple(self.results['interactor_name'].unique())

    def unique_drugs(self):
        """Return a list of unique drugs found in the results dataframe."""
        if self.results is not None:
            return tuple(self.results['drug'].unique())


def get_interactor_list(results_df: pd.DataFrame):
    """Collect interactors from InteractorFinder results."""
    interactors = set()
    for gene_list in results_df.interactor_involved_genes:
        if gene_list is not None:
            for gene in gene_list:
                interactors.add(gene)
    for other_list in results_df.interactor_involved_other:
        if other_list is not None:
            for other in other_list:
                interactors.add(other)
    for name in results_df.interactor_name:
        interactors.add(name)

    return interactors



"""Table definitions for SQLite DB."""

from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, VARCHAR, INTEGER, BOOLEAN, ForeignKey

from drugintfinder.defaults import engine

Base = declarative_base()


class MetaClass:
    """Meta methods for handling DB tables."""

    def dump(self):
        """Delete all tables in DB."""
        return dict([(k, v) for k, v in vars(self).items() if not k.startswith('_')])


class Druggable(MetaClass, Base):
    """Druggable interactor table definitions."""

    __tablename__ = 'druggable_interactor_relations'
    id = Column(Integer, primary_key=True)

    drug = Column(VARCHAR(1000))
    drugbank_id = Column(VARCHAR(255))
    chembl_id = Column(VARCHAR(255))
    pubchem_id = Column(VARCHAR(255))
    interactor_type = Column(VARCHAR(100))
    interactor_bel = Column(VARCHAR(1000))
    interactor_name = Column(VARCHAR(255))
    capsule_interactor_bel = Column(VARCHAR(1000))
    capsule_interactor_type = Column(VARCHAR(255))
    rel_pub_year = Column(INTEGER)
    target_bel = Column(VARCHAR(1000))
    target_symbol = Column(VARCHAR(255), index=True)
    target_type = Column(VARCHAR(255), index=True)
    relation_type = Column(VARCHAR(255))
    pmod_type = Column(VARCHAR(255))
    pmid = Column(INTEGER)
    pmc = Column(VARCHAR(255))
    rel_rid = Column(VARCHAR(255))
    drug_rel_rid = Column(VARCHAR(255))
    drug_rel_actions = Column(VARCHAR(255))


class General(MetaClass, Base):
    """General table definitions."""

    __tablename__ = 'general'
    id = Column(Integer, primary_key=True)

    pmod_type = Column(VARCHAR(255))
    target_bel = Column(VARCHAR(1000))
    target_symbol = Column(VARCHAR(255), index=True)
    target_type = Column(VARCHAR(255), index=True)
    relation_type = Column(VARCHAR(255))
    interactor_bel = Column(VARCHAR(1000))
    interactor_name = Column(VARCHAR(255))
    interactor_type = Column(VARCHAR(255))
    pmid = Column(INTEGER)
    pmc = Column(VARCHAR(255))
    target_species = Column(INTEGER)


class Patents(MetaClass, Base):
    """Patents table definitions."""

    __tablename__ = 'patents'
    id = Column(Integer, primary_key=True)

    expired = Column(BOOLEAN)
    patent_number = Column(Integer)
    drug_id = Column(Integer, ForeignKey('drugs.id'), index=True)


class Products(MetaClass, Base):
    """Products table definitions."""

    __tablename__ = 'products'
    id = Column(Integer, primary_key=True)

    product_name = Column(VARCHAR(255), index=True)
    has_generic = Column(BOOLEAN)
    is_approved = Column(BOOLEAN)
    has_approved_generic = Column(BOOLEAN)
    drug_id = Column(Integer, ForeignKey('drugs.id'), index=True)


class Trials(MetaClass, Base):
    """Clinical trials table definitions."""

    __tablename__ = 'trials'
    id = Column(Integer, primary_key=True, index=True)

    trial_id = Column(VARCHAR(255), index=True, unique=True)
    trial_status = Column(VARCHAR(255))
    conditions = Column(VARCHAR(255))
    drugs_in_trial = Column(VARCHAR(255))


class TargetStats(MetaClass, Base):
    """Target statistic table definitions."""

    __tablename__ = 'target_stats'
    id = Column(Integer, primary_key=True)

    symbol = Column(VARCHAR(255), index=True)
    out_count = Column(INTEGER)
    in_count = Column(INTEGER)
    both_count = Column(INTEGER)


class Drugs(MetaClass, Base):
    """Drugs table definitions."""

    __tablename__ = 'drugs'
    id = Column(Integer, primary_key=True, index=True)

    drug_name = Column(VARCHAR(255), index=True)
    drugbank_id = Column(VARCHAR(255), index=True, unique=True)
    num_targets = Column(INTEGER)
    targets = Column(VARCHAR(255))
    patents = relationship("Patents")
    products = relationship("Products")
    clinical_trials = Column(VARCHAR(255))


class BioAssays(MetaClass, Base):
    """BioAssay table definitions."""

    __tablename__ = 'bioassays'
    id = Column(Integer, primary_key=True)

    target = Column(VARCHAR(255), index=True, unique=True)
    num_assays = Column(INTEGER)


Base.metadata.create_all(engine)
